/**
 **	File ......... Socks4Socket.cpp
 **	Published ....  2005-03-23
 **	Author ....... grymse@alhem.net
**/
/*
Copyright (C) 2004  Anders Hedstrom

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
//#include <stdio.h>
#include <ListenSocket.h>

#include "Socks4Handler.h"
#include "RemoteSocket.h"
#include "Socks4Socket.h"




Socks4Socket::Socks4Socket(SocketHandler& h)
:TcpSocket(h)
,m_state(STATE_VN)
,m_userid_ptr(0)
,m_pRemote(NULL)
{
}


Socks4Socket::~Socks4Socket()
{
}




void Socks4Socket::OnRead()
{
/*
		STATE_VN = 0,
		STATE_CD,
		STATE_DSTPORT,
		STATE_DSTIP,
		STATE_USERID,
		STATE_TRY_CONNECT,
		STATE_CONNECTED,
		STATE_FAILED,
*/
	// ibuf is inherited member variable of TcpSocket
	TcpSocket::OnRead();
	bool need_more = false;
	while (ibuf.GetLength() && !need_more && !CloseAndDelete() )
	{
		size_t l = ibuf.GetLength();
		switch (m_state)
		{
		case STATE_VN: // get socks version
			ibuf.Read( (char *)&m_vn, 1);
			if (m_vn != 4)
			{
				Handler().LogError(this, "OnRead", m_vn, "Unsupported socks version", LOG_LEVEL_FATAL);
				SetCloseAndDelete();
			}
			else
			{
				m_state = STATE_CD;
			}
			break;
		case STATE_CD:
			ibuf.Read( (char *)&m_cd, 1);
			if (m_cd != 1 && m_cd != 2)
			{
				Handler().LogError(this, "OnRead", m_cd, "Bad command code", LOG_LEVEL_FATAL);
				SetCloseAndDelete();
			}
			else
			{
				m_state = STATE_DSTPORT;
			}
			break;
		case STATE_DSTPORT:
			if (l > 1)
			{
				ibuf.Read( (char *)&m_dstport, 2);
				m_state = STATE_DSTIP;
			}
			else
			{
				need_more = true;
			}
			break;
		case STATE_DSTIP:
			if (l > 3)
			{
				ibuf.Read( (char *)&m_dstip, 4);
				m_state = STATE_USERID;
			}
			else
			{
				need_more = true;
			}
			break;
		case STATE_USERID:
			if (m_userid_ptr < MAXLEN_USERID)
			{
				ibuf.Read(m_userid + m_userid_ptr, 1);
				if (m_userid[m_userid_ptr] == 0)
				{
					switch (m_cd)
					{
					case 1:
						SetupConnect();
						break;
					case 2:
						SetupBind();
						break;
					}
				}
				m_userid_ptr++;
			}
			else
			{
				Handler().LogError(this, "OnRead", m_userid_ptr, "Userid too long", LOG_LEVEL_FATAL);
				SetCloseAndDelete();
			}
			break;
		case STATE_TRY_CONNECT:
			break;
		case STATE_CONNECTED:
			{
				char buf[TCP_BUFSIZE_READ];
				ibuf.Read(buf, l);
				if (m_pRemote && Handler().Valid(m_pRemote))
				{
					m_pRemote -> SendBuf(buf, l);
				}
			}
			break;
		case STATE_FAILED:
			SetCloseAndDelete();
			break;
		}
	}
}


void Socks4Socket::Connected()
{
	if (m_state == STATE_TRY_CONNECT)
	{
		char reply[8];
		reply[0] = 0;
		reply[1] = 90; // granted
		memcpy(reply + 2, &m_dstport, 2);
		memcpy(reply + 4, &m_dstip, 4);
		SendBuf(reply, 8);
		m_state = STATE_CONNECTED;
	}
	else
	{
		Handler().LogError(this, "Connected", m_state, "bad state", LOG_LEVEL_FATAL);
		SetCloseAndDelete();
	}
}


void Socks4Socket::ConnectFailed()
{
	if (m_state == STATE_TRY_CONNECT)
	{
		char reply[8];
		reply[0] = 0;
		reply[1] = 91; // failed
		memcpy(reply + 2, &m_dstport, 2);
		memcpy(reply + 4, &m_dstip, 4);
		SendBuf(reply, 8);
		m_state = STATE_FAILED;
		SetCloseAndDelete();
	}
	else
	{
		Handler().LogError(this, "ConnectFailed", m_state, "bad state", LOG_LEVEL_FATAL);
		SetCloseAndDelete();
	}
}


void Socks4Socket::SetupConnect()
{
	m_pRemote = new RemoteSocket(Handler());
	port_t port = ntohs(m_dstport);
	ipaddr_t addr = m_dstip; //ntohl(m_dstip);
	m_pRemote -> Open(addr, port);
	m_pRemote -> SetDeleteByHandler();
	m_pRemote -> SetRemote(this);
	Handler().Add(m_pRemote);
	//
	m_state = STATE_TRY_CONNECT;
}


void Socks4Socket::SetupBind()
{
	ListenSocket<RemoteSocket> *p = new ListenSocket<RemoteSocket>(Handler());
	p -> SetDeleteByHandler();
	p -> Bind();
	Handler().Add(p);
	static_cast<Socks4Handler&>(Handler()).SetBind(p, this);

	char reply[8];
	reply[0] = 0;
	reply[1] = 90;
	port_t port = p -> GetPort();
	unsigned short s = htons(port);
	memcpy(reply + 2,&s,2);
	memset(reply + 4,0,4);
	SendBuf(reply, 8);
	//
	m_state = STATE_ACCEPT;
}


void Socks4Socket::Accept()
{
	if (m_state == STATE_ACCEPT)
	{
		ipaddr_t remote = m_pRemote -> GetRemoteIP4();
		ipaddr_t dstip = ntohl(m_dstip);
		char reply[8];
		reply[0] = 0;
		reply[1] = (dstip == remote) ? 90 : 91;
		memcpy(reply + 2, &m_dstport, 2);
		memcpy(reply + 4, &m_dstip, 4);
		SendBuf(reply, 8);
		m_state = (dstip == remote) ? STATE_CONNECTED : STATE_FAILED;
		if (m_state == STATE_FAILED)
		{
			Handler().LogError(this, "Accept", remote, "Remote ip mismatch", LOG_LEVEL_FATAL);
			SetCloseAndDelete();
		}
	}
	else
	{
		Handler().LogError(this, "Accept", m_state, "bad state", LOG_LEVEL_FATAL);
		SetCloseAndDelete();
	}
}


